"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from cahoots.parsers.base import BaseParser
from cahoots.parsers.programming.lexer import ProgrammingLexer
from cahoots.parsers.programming.bayesian import ProgrammingBayesianClassifier
from cahoots.result import ParseResult
from SereneRegistry import registry
import os
import re
import glob
import yaml


class ProgrammingParser(BaseParser):
    """Determines if a sample of text is a programming language"""

    all_keywords = []
    language_keywords = {}

    @staticmethod
    def bootstrap(config):
        """Loads tokens from the yaml files on disk"""
        all_keywords = []
        language_keywords = {}

        directory = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(directory, "languages/*.yaml")

        for file_path in glob.glob(path):
            with open(file_path, 'r') as language_file:
                language = yaml.load(language_file)
                all_keywords.extend(language['keywords'])
                language_keywords[language['id']] = language

        registry.set('PP_all_keywords', set(all_keywords))
        registry.set('PP_language_keywords', language_keywords)

        ProgrammingBayesianClassifier.bootstrap(config)

    def __init__(self, config):
        BaseParser.__init__(self, config, "Programming", 0)
        self.all_keywords = registry.get('PP_all_keywords')
        self.language_keywords = registry.get('PP_language_keywords')

    # finding common words/phrases in programming languages
    def find_common_tokens(self, dataset):
        """Looking for any programming language keywords in the data"""
        return [keyword for keyword in dataset if keyword in self.all_keywords]

    @classmethod
    def basic_language_heuristic(cls, language_data, dataset):
        """Looking for specific programming language keywords in the data"""
        return [keyword for keyword in
                dataset if keyword in language_data['keywords']]

    @classmethod
    def create_dataset(cls, data):
        """Takes a data string and converts it into a dataset for analysis"""
        return set(re.split('[ ;,{}()\n\t\r]', data.lower()))

    @classmethod
    def normalize_scores(cls, data, lex_languages, bayes_languages):
        """Normalizing the parsing scores generated by lex and bayes parsing"""

        # Giving ourselves a maximum of 10% confidence
        # addition for lexer detection
        normalizer = 10 / float(
            max([scr for _, scr in lex_languages.items()])
        )
        normal_scores = {}

        for lang_id, score in lex_languages.items():
            normal_scores[lang_id] = (normalizer * score)

        # Pulling some stats out of our bayes
        # results so we can calculate a confidence
        results = [[lid, scr] for lid, scr in
                   bayes_languages.items() if lid in lex_languages]
        scores = [scr for lid, scr in results]
        min_score = min(scores)
        spread = max(scores) - min_score

        # We want to add up to 90% confidence based on the spread from min
        # to max matches Math for the actual normalization is explained in
        # a comment below
        if spread > 90:
            normalizer = 90.00 / spread
        else:
            normalizer = 1.00

        # Normalizing and assembling results
        for lang_id, score in results:

            # Example:
            #
            #     min_score = -1700
            #     score = -1500
            #     Therefore: score = -1500 + 1700 (200)
            #
            #     OR
            #
            #     min_score = -50
            #     score = -25
            #     Therefore: score = 25
            if min_score < -90:
                score = score + (-1 * min_score)
            else:
                score = -1 * score

            # Example:
            #     spread > 90 example:
            #         score = 210
            #         spread = 240 (ie: max of -1700 to min of -1940)
            #         normalizer = 90 / 240 = 0.375
            #         normalized score = 78.75
            #
            #     spread <= 90 example:
            #         score 25
            #         spread = 30
            #         normalizer = 1
            #         normalized score = 25
            normal_scores[lang_id] += (normalizer * score)

        # We reduce our confidence significantly
        # if the string provided is very short
        normalizer = (min(100, float(len(data))) / 100)
        if normalizer < 1:
            for lang_id in normal_scores:
                normal_scores[lang_id] = normal_scores[lang_id] * normalizer

        # turning the score into an integer
        for lang_id in normal_scores:
            normal_scores[lang_id] = int(round(normal_scores[lang_id]))

        return normal_scores

    def get_possible_languages(self, dataset):
        """
        Goes through each language's keywords and finds
        languages that share keywords with the provided data set
        """
        return [language for language, language_data in
                self.language_keywords.items() if
                self.basic_language_heuristic(
                    language_data, dataset
                )]

    def parse(self, data):
        """
        Determines if the data is an example of one of our trained languages
        """

        dataset = self.create_dataset(data)

        # Step 1: Is this possibly code?
        if not self.find_common_tokens(dataset):
            return

        # Step 2: Which languages match, based on keywords alone?
        matched_languages = self.get_possible_languages(dataset)

        # Step 3: Which languages match, based on a smarter lexer?
        lexer = ProgrammingLexer(matched_languages, data.lower())
        lex_languages = lexer.lex()

        if not lex_languages:
            return

        # Step 4: Using a Naive Bayes Classifier
        # to pinpoint the best language fits
        classifier = ProgrammingBayesianClassifier()
        bayes_languages = classifier.classify(data)

        scores = self.normalize_scores(data, lex_languages, bayes_languages)

        for lang_id, confidence in scores.items():
            yield ParseResult(self.type,
                              self.language_keywords[lang_id]['name'],
                              confidence)



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from cahoots.parsers.base import BaseParser
from cahoots.parsers.name import NameParser
from cahoots.parsers.location import \
    LocationDatabase, StreetSuffixEntity
from SereneRegistry import registry
import re
import sqlite3
import itertools


class AddressParser(BaseParser):
    """
    Determines if given data is an address

    We don't use regex to find addresses because it's pretty much
    impossible to match all the possible (infinite) combinations.

    Instead, we look for elements that addresses have, or don't have,
    and determine if it's likely that this is an address or not.
    """

    subtype = "Without Addressee"

    def __init__(self, config):
        BaseParser.__init__(self, config, "Address", 0)

    @staticmethod
    def bootstrap(config):
        """preps the address parser"""
        split_regex = re.compile(r"[\s#`'?.;,-/]")
        registry.set('AP_split_regex', split_regex)

    @classmethod
    def get_street_suffix(cls, data_set):
        """Looks at our data for a street suffix and returns the entity"""
        sql_items = []

        for _ in data_set:
            sql_items.append('suffix_name = ?')

        database = LocationDatabase.get_database()
        cursor = database.cursor()

        try:
            row = cursor.execute(
                'SELECT * FROM street_suffix WHERE ' + " or ".join(sql_items),
                tuple(data_set)
            ).fetchone()
            database.close()
        except sqlite3.Error:
            database.close()
            return None

        if row is not None:
            return LocationDatabase.hydrate(row, StreetSuffixEntity)

    @classmethod
    def separate_suspected_name(cls, data_set):
        """Separates a suspected name from the beginning of an address"""
        name_elems = []
        work_data = list(data_set)
        conjunctions = ['and', 'but', 'or', 'yet', 'for', 'nor', 'so']
        for word in data_set:
            if not word[0].isdigit():
                work_data.pop(0)
                # skipping conjunctions, ex: Alice and Bob Saget
                if word.lower() not in conjunctions:
                    name_elems.append(word)
            else:
                break

        return (list(work_data), " ".join(name_elems))

    def is_address_name(self, suspected_name):
        """Checking if the start of an address bears a name"""
        name_parser = NameParser(self.config)
        for _ in name_parser.parse(suspected_name):
            return True
        return False

    @classmethod
    def is_invalid_int(cls, word):
        """Looking for ints that are to short (less than 5 chars)"""
        try:
            int(word)
            if len(word) < 5:
                return True
            else:
                return False
        except ValueError:
            return False

    def find_address_tokens(self, data, results):
        """Looking for known city or country, zip, etc."""
        tokens = 0
        working_data = list(data)

        data_combinations = list(itertools.combinations(working_data, 2))
        data_combinations = [" ".join(x) for x in data_combinations]

        working_data.extend(data_combinations)

        database = LocationDatabase.get_database()
        cursor = database.cursor()

        for test in [x for x in working_data if not self.is_invalid_int(x)]:

            params = (test, test, test, test, test)

            sql = 'SELECT count(*) FROM city WHERE ' +\
                  'country = ? OR ' +\
                  'postal_code = ? OR ' +\
                  'city = ? OR ' +\
                  'state1 = ? OR ' +\
                  'state2 = ?'

            try:
                row = cursor.execute(sql, params).fetchone()
                row = row[0]
            except sqlite3.Error:
                database.close()
                return 0, results

            if row:
                entity = {}
                entity['token'] = test
                entity['matches'] = row
                results['address_tokens'].append(entity)
                tokens += 1
                self.confidence += 20

        database.close()

        return (tokens, results)

    def generate_result_data(self, data, data_set):
        """
        Examines the data_set and generates a result
        dict, or returns nothing if there's a failure
        """
        results = {
            'street_suffix': None,
            'addressed_to': None,
            'address_tokens': []
        }

        # With no street suffix, we don't consider this an address
        suffix = self.get_street_suffix(data_set)
        if not suffix:
            return

        suffix_test = [x.lower() for x in data_set]
        try:
            # removing the suffix from the list of working data
            index = suffix_test.index(suffix.suffix_name)
            data_set.pop(index)
        except ValueError:
            # if for some reason we found an invalid suffix, we probably cry
            return

        results['street_suffix'] = suffix.suffix_name

        # Looking to see if this address starts with a name, since no digit
        if not data[0].isdigit():
            data_set, suspected_name = \
                self.separate_suspected_name(data_set)

            results['addressed_to'] = suspected_name

            if NameParser in self.config.enabled_modules:
                if self.is_address_name(suspected_name):
                    self.subtype = "With Addressee"
                    self.confidence += 25
                else:
                    # If it starts with non-name text, not an address
                    return

        # Looking for items in the city database that match words in the data
        token_count, results = self.find_address_tokens(data_set, results)

        return (results, token_count, data_set)

    def parse(self, data):
        """parses for address"""
        data = data.strip()

        # If invalid length or there are no digits, we return.
        if len(data) > 100 or not [x for x in data if x.isdigit()]:
            return

        split_regex = registry.get('AP_split_regex')
        # splitting the data string and removing empty values
        data_set = [x for x in split_regex.split(data) if x]

        # At least 4 words and one of the words should start with a number
        if len(data_set) <= 3 or not [x for x in data_set if x[:1].isdigit()]:
            return

        results, token_count, data_set = \
            self.generate_result_data(data, data_set) \
            or (None, None, None)

        if token_count:
            # Subtracting a little confidence for each token that wasn't found
            self.confidence -= 5*(len(data_set)-token_count)

            yield self.result(self.subtype, min(100, self.confidence), results)



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from cahoots.parsers.base import BaseParser
from SereneRegistry import registry
from cahoots.parsers.location import \
    LocationDatabase, CityEntity
import re
import sqlite3


class PostalCodeParser(BaseParser):
    """Detects if the data provided is a location"""

    def __init__(self, config):
        BaseParser.__init__(self, config, "Postal Code", 100)

    @staticmethod
    def bootstrap(config):
        """Bootstraps the location parser"""
        # Will test if something matches 5 or 9 digit postalcode pattern
        postal_regex = re.compile(
            r'^' +
            r'(\d{2,7}(-\d{2,4})?)|' +
            r'([a-zA-Z]\d{3})|' +
            r'([a-zA-Z]{2}\s\d{2})|' +
            r'([a-zA-Z]{2}-\d{2})|' +
            r'(AD\d{3})|' +
            r'(\d{3}\s\d{2})|' +
            r'([a-zA-Z]{2}\d{4})|' +
            r'(\d{4}\sW3)|' +
            r'(\d{4}\s[a-zA-Z]{2})|' +
            r'([a-zA-Z]\d[a-zA-Z]\s\d[a-zA-Z]\d)|' +
            r'(AZ\s\d{4})|' +
            r'(BB\d{1,5})|' +
            r'([a-zA-Z]{2}\d{1,2}\s\d[a-zA-Z]{2})|' +
            r'(JMA[a-zA-Z]{2}\d{2})|' +
            r'(AZ-\d{4})|' +
            r'([a-zA-Z]\d{4}[a-zA-Z]{3})|' +
            r'([a-zA-Z]{2}\d{2}\s\d[a-zA-Z]{2})|' +
            r'([a-zA-Z]{3}\s\d{4})|' +
            r'([a-zA-Z]{4}\s1ZZ)|' +
            r'([a-zA-Z]{2}\d{1,2}(-\d{4})?)|' +
            r'(\d{5}\sCEDEX(\s\d{1,2})?)' +
            r'$'
        )
        registry.set('ZCP_postal_code_regex', postal_regex)

    @classmethod
    def get_postal_code_data(cls, data):
        """If this looks like a postal code, we try to get its info"""
        plus_postal_code = '-' in data

        if plus_postal_code:
            prefix = data.split('-')[0]

        database = LocationDatabase.get_database()
        cursor = database.cursor()

        try:
            rows = cursor.execute(
                'SELECT * FROM city WHERE postal_code = ?',
                (data,)
            ).fetchall()
            entities = LocationDatabase.hydrate(rows, CityEntity)
        except sqlite3.Error:
            database.close()
            return None

        if plus_postal_code:
            try:
                rows = cursor.execute(
                    'SELECT * FROM city WHERE postal_code = ?',
                    (prefix,)
                ).fetchall()
                prefix_entities = LocationDatabase.hydrate(rows, CityEntity)
                entities.extend(prefix_entities)
            except sqlite3.Error:
                database.close()
                return None

        if not entities:
            database.close()
            return None

        entities = LocationDatabase.substitute_country_data(entities, cursor)
        entities = [vars(x) for x in entities]

        database.close()
        return entities

    def calculate_confidence(self, data, results):
        """calculates the confidence that this is a postal code"""

        # The longer the data string, the higher the confidence
        self.confidence -= (40-(2*len(data)))

        if len(results) > 1:
            self.confidence -= (7 * len(results))

    def parse(self, data):
        """parses data to determine if this is a location"""
        data = data.strip()

        if len(data) >= 20:
            return

        postal_regex = registry.get('ZCP_postal_code_regex')
        if postal_regex.match(data):
            results = self.get_postal_code_data(data)
            if results is not None:
                self.calculate_confidence(data, results)
                if self.confidence > 0:
                    yield self.result("Postal Code", self.confidence, results)
                    return



# coding: utf-8
"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
# pylint: disable=anomalous-backslash-in-string
from cahoots.parsers.base import BaseParser
from SereneRegistry import registry
from LatLon23 import string2latlon, LatLon
from cahoots.util import u
import re


class CoordinateParser(BaseParser):
    """Detects if the data provided is a coordinate"""

    def __init__(self, config):
        BaseParser.__init__(self, config, "Coordinates", 100)

    @staticmethod
    def bootstrap(config):
        """Bootstraps the coordinate parser"""
        # Will test if something matches regular coordinates
        # 34.56,23.65 or 34.56 23.65 or 34.56 , 23.65
        coord_regex = re.compile(
            r'^(-?\d{1,3}(?:\.\d+)?)' +
            r'(?:(?:(?:\s+)?,(?:\s+)?)|(?:\s+))' +
            r'(-?\d{1,3}(?:\.\d+))?$'
        )
        registry.set('CP_coord_regex', coord_regex)

        # Will test if something matches degree coordinates
        # 40.244° N 79.123° W
        deg_regex = re.compile(
            u('^(\d{1,3}\.\d+°?\s+[nNsS])') +
            u('\s+') +
            u('(\d{1,3}\.\d+°?\s+[wWeE])$')
        )
        registry.set('CP_deg_regex', deg_regex)

        # Will test if something matches deg/min coordinates
        # 13° 34.425' N 45° 37.983' W
        deg_min_regex = re.compile(
            u('^(\d{1,3}°?\s+\d{1,3}\.\d+\'?\s+[nNsS])') +
            u('\s+') +
            u('(\d{1,3}°?\s+\d{1,3}\.\d+\'?\s+[wWeE])$')
        )
        registry.set('CP_deg_min_regex', deg_min_regex)

        # Will test if something matches deg/min/sec coordinates
        # 40° 26' 46.56" N 79° 58' 56.88" W
        deg_min_sec_regex = re.compile(
            u('^(\d{1,3}°?\s+\d{1,3}\'?\s+\d{1,3}(?:\.\d+)?"?\s+[nNsS])') +
            u('\s+') +
            u('(\d{1,3}°?\s+\d{1,3}\'?\s+\d{1,3}(?:\.\d+)?"?\s+[wWeE])$')
        )
        registry.set('CP_deg_min_sec_regex', deg_min_sec_regex)

    @classmethod
    def clean_dms_coords(cls, coord_string):
        """Removes items that the LatLon parser doesn't want"""
        coord_string = coord_string.replace(u('°'), '')
        coord_string = coord_string.replace('\'', '')
        coord_string = coord_string.replace('"', '')
        return coord_string

    @classmethod
    def get_standard_coord_data(cls, match):
        """Creates LatLon from regular coords"""
        lat_lon = LatLon(
            float(match.group(1)),
            float(match.group(2))
        )
        return lat_lon

    def get_degree_based_coord_data(self, match, pattern):
        """Takes degree based coords and turns them into LatLon object"""
        lat_lon = string2latlon(
            self.clean_dms_coords(match.group(1)),
            self.clean_dms_coords(match.group(2)),
            pattern
        )
        return lat_lon

    @classmethod
    def generate_result_data(cls, result):
        """Adds a google map link to the result data"""
        result_dict = {
            'latitude': result[0],
            'longitude': result[1]
        }

        # throwing this in there for funsies
        url = 'https://www.google.com/maps?q=' + result[0] + ',' + result[1]
        additional_data = {
            'map_url': url
        }

        return result_dict, additional_data

    def get_coordinate_test_parameters(self):
        """Returns the test parameters for coordinate tests"""
        return [
            (
                # Standard coordinate match
                'CP_coord_regex',  # Name of the registry key holding regex
                self.get_standard_coord_data,  # formatting function
                None,  # Specific input format for the formatting function
                'Standard',  # Subtype if match
                80,  # Confidence if match
            ),
            (
                # Degree coordinate match
                'CP_deg_regex',
                self.get_degree_based_coord_data,
                'd% %H',
                'Degree',
                100,
            ),
            (
                # Degree/minute coordinate match
                'CP_deg_min_regex',
                self.get_degree_based_coord_data,
                'd% %m% %H',
                'Degree/Minute',
                100,
            ),
            (
                # Degree/minutes/second coordinate match
                'CP_deg_min_sec_regex',
                self.get_degree_based_coord_data,
                'd% %m% %S% %H',
                'Degree/Minute/Second',
                100,
            ),
        ]

    def parse(self, data):
        """parses data to determine if this is a location"""
        data = data.strip()

        test_parameters = self.get_coordinate_test_parameters()

        for reg_key, format_func, fmt, subtype, confidence in test_parameters:
            # checking each of our types of coordinates and breaking on find
            match = registry.get(reg_key).match(data)
            if match:
                # if a format_arg provided, we pass it into formatting func
                res = format_func(match, fmt) if fmt else format_func(match)

                # Prepping processed data with better metadata
                result, add_data = self.generate_result_data(res.to_string())
                yield self.result(subtype, confidence, result, add_data)

                # Only looking to match one format, so we break here
                break



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
# Cahoots Parsers
from cahoots.parsers.boolean import BooleanParser
from cahoots.parsers.character import CharacterParser
from cahoots.parsers.date import DateParser
from cahoots.parsers.email import EmailParser
from cahoots.parsers.equation import EquationParser
from cahoots.parsers.measurement import MeasurementParser
from cahoots.parsers.name import NameParser
from cahoots.parsers.number import NumberParser
from cahoots.parsers.phone import PhoneParser
from cahoots.parsers.programming import ProgrammingParser
from cahoots.parsers.uri import URIParser
from cahoots.parsers.location.address import AddressParser
from cahoots.parsers.location.coordinate import CoordinateParser
from cahoots.parsers.location.landmark import LandmarkParser
from cahoots.parsers.location.postalcode import PostalCodeParser
# Hierarchical Confidence Normalizers
from cahoots.confidence.normalizers.character import CharacterWithoutBoolean
from cahoots.confidence.normalizers.date import DateWithPostalCode
from cahoots.confidence.normalizers.equation import \
    EquationWithPhonePostalCode
from cahoots.confidence.normalizers.number import \
    IntOctWithPhoneDatePostalCode, NumberWithNonNumbers
from cahoots.confidence.normalizers.phone import PhoneWithUri


class BaseConfig(object):
    """
    Cahoots Configuration
    Change this file to suit your installation's needs.
    """

    # Are we in debug mode?
    debug = False

    """
    To disable a module, simply comment it out of this list.
        Be aware that modules may have soft-dependencies on one another,
        and a disabled module may impact results results from other
        modules which seek its "council." This may result in unexpected
        confidence scores, a change in result determination, etc.
    """
    enabled_modules = [
        AddressParser,
        BooleanParser,
        CharacterParser,
        CoordinateParser,
        DateParser,
        EmailParser,
        EquationParser,
        LandmarkParser,
        MeasurementParser,
        NameParser,
        NumberParser,
        PhoneParser,
        PostalCodeParser,
        ProgrammingParser,
        URIParser,
    ]

    enabled_confidence_normalizers = [
        CharacterWithoutBoolean,
        DateWithPostalCode,
        EquationWithPhonePostalCode,
        IntOctWithPhoneDatePostalCode,
        NumberWithNonNumbers,
        PhoneWithUri,
    ]



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from cahoots.parsers.base import BaseParser
from cahoots.parsers.number import NumberParser
import re
import string


class NameParser(BaseParser):
    '''Determines if given data is a name'''

    prefixes = ['MS', 'MISS', 'MRS', 'MR', 'MASTER', 'REV', 'FR', 'DR',
                'ATTY', 'PROF', 'HON', 'PRES', 'GOV', 'COACH', 'OFC', 'MSGR',
                'SR', 'BR', 'SUPT', 'REP', 'SEN', 'AMB', 'TREAS', 'SEC',
                'PVT', 'CPL', 'SGT', 'ADM', 'MAJ', 'CAPT', 'CMDR', 'LT',
                'COL', 'GEN']
    suffixes = ['AB', 'BA', 'BFA', 'BTECH', 'LLB', 'BSC', 'MA', 'MFA', 'LLM',
                'MLA', 'MBA', 'MSC', 'JD', 'MD', 'DO', 'PHARMD', 'PHD',
                'DPHIL', 'LLD', 'ENGD', 'KBE', 'LLD', 'DD', 'ESQ', 'ESQUIRE',
                'CSA', 'ASCAP', 'CA', 'CPA', 'CFA', 'PE', 'PG', 'CPL', 'CENG',
                'RA', 'AIA', 'RIBA', 'MEOA', 'CISA', 'CISSP', 'CISM', 'PT',
                'DPT', 'MCSP', 'SRP', 'RGN', 'USN', 'USMC', 'OFM', 'CSV',
                'JR', 'SR', 'JNR', 'SNR', '2ND', '3RD', '4TH', '5TH', '6TH',
                '7TH', '8TH', '9TH', 'BT', 'BART', 'QC', 'MP', 'SSF', 'FRCP',
                'FRSA', 'RAF', 'RN', 'RMP', 'FAIA', 'FRSE', 'SJ', 'OP',
                'ICMA-CM', 'MBASW']

    def __init__(self, config):
        BaseParser.__init__(self, config, "Name", 0)

    @classmethod
    def basic_validation(cls, data):
        """
        Make sure every word in the phrase either
        starts with a Capital Letter or a Number
        """
        return len(data) == len(
            [word for word in data if
             # Has to start with a capital letter or digit
             (word[0].isupper() or (len(data) > 1 and word[0].isdigit())) and
             # Whole word can't be a digit
             not word.isdigit() and
             # Must contain only printable characters
             not [char for char in word if char not in string.printable]]
        )

    def is_prefix(self, word):
        """Checks to see if the word passed in is a name prefix"""
        word = word.replace('.', '').upper()
        return word in self.prefixes

    def is_suffix(self, word):
        """Checks to see if the word passed in is a name suffix"""
        if NumberParser in self.config.enabled_modules:
            nump = NumberParser(self.config)
            if nump.is_roman_numeral(word) != (False, 0):
                return True

        word = word.replace('.', '').upper()
        return word in self.suffixes

    @classmethod
    def is_initial(cls, word):
        """Checks to see if the word passed in is an initial"""
        if len(word) > 2:
            return False

        if len(word) == 1:
            return word[0].isalpha()

        if len(word) == 2:
            return word[1] == '.' and word[0].isalpha()

    def detect_prefix_or_suffix(self, data):
        """
        Checking for things like Mr. or Jr. Big boost for these values.
        If found, we remove them from the list of words
        """
        if len(data) > 2:
            if self.is_prefix(data[0]):
                data.pop(0)
                self.confidence += 20
            if self.is_suffix(data[-1]):
                data.pop()
                self.confidence += 20

    def calculate_confidence(self, data):
        '''Calculates confidence based on various attributes of the data'''

        # If we have a two - four word "name" here we boost its
        # confidence since this is something of a giveaway
        if len(data) in range(2, 4):
            self.confidence += (5 * len(data))

        # Adding confidence for initials vs words,
        # if we have greater than one word
        if len(data) > 1:
            for word in data:
                if self.is_initial(word):
                    self.confidence += 15
                    # if there's a period in this initial, we boost it.
                    if "." in word:
                        self.confidence += 5
                else:
                    self.confidence += 10

        # If our "name" is longer than 4 words, we
        # reduce the likelihood that it's a name
        if len(data) > 3:
            self.confidence -= (7*len(data))

    def parse(self, data):
        """Determines if the data is a name or not"""

        # Making sure there are at least SOME uppercase letters in the phrase
        if not re.search('[A-Z]', data):
            return

        data = data.split()

        # If someone has a name longer than 7 words...they need
        # help. Making sure each word in the phrase starts with an
        # uppercase letter or a number
        if len(data) >= 7 or not self.basic_validation(data):
            return

        self.detect_prefix_or_suffix(data)

        self.calculate_confidence(data)

        if self.confidence <= 0:
            return

        yield self.result("Name", min(100, self.confidence))



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import sqlite3
import os


class LocationDatabase(object):
    """Handles the database for location parsing in various submodules"""

    @classmethod
    def get_database(cls):
        """Gets a sqlite database connection to the location database"""
        data_dir = os.path.dirname(os.path.realpath(__file__))
        database = sqlite3.connect(data_dir + '/data/location.sqlite')
        database.cursor().execute('PRAGMA temp_store = 2')
        return database

    def single_select(self, sql, params, prototype):
        """Executes a statement (assumed to be a select) against the db"""
        if not isinstance(params, tuple):
            raise TypeError(
                'LocationDatabase.select requires params to be a tuple.'
            )

        database = self.get_database()
        cursor = database.cursor()

        try:
            rows = cursor.execute(sql, params).fetchall()
        except sqlite3.Error:
            database.close()
            return None

        database.close()

        if len(rows) > 0:
            return self.hydrate(rows, prototype)
        else:
            return None

    @classmethod
    def hydrate(cls, data, prototype):
        """hydrates a set of entities"""
        if isinstance(data, list):
            entities = []
            for row in data:
                entities.append(prototype(row))
            return entities
        elif isinstance(data, tuple):
            return prototype(data)
        else:
            raise TypeError(
                'Location entity hydration requires a list or tuple.'
            )

    @classmethod
    def substitute_country_data(cls, entities, cursor):
        """Preps our entity objects with country data and converts dicts"""
        results = []

        for result in entities:
            try:
                rows = cursor.execute(
                    'SELECT * FROM country WHERE abbreviation = ?',
                    (result.country,)
                ).fetchall()
            except sqlite3.Error:
                rows = []

            if rows:
                entities = LocationDatabase.hydrate(rows, CountryEntity)
                entity = entities[0]
                entity.name = entity.name.title()
                entity.abbreviation = entity.abbreviation.upper()
                result.country = vars(entity)

            results.append(result)

        return results


# pylint: disable=too-many-instance-attributes
class CityEntity(object):
    """Represents a city database entity"""

    def __init__(self, data):
        self.country,\
            self.postal_code,\
            self.city,\
            self.state1,\
            self.state2,\
            self.province1,\
            self.province2,\
            self.community1,\
            self.community2,\
            self.latitude,\
            self.longitude,\
            self.coord_accuracy = data


class LandmarkEntity(object):
    """Represents a landmark database entity"""

    def __init__(self, data):
        self.resource,\
            self.address,\
            self.city,\
            self.county,\
            self.state,\
            self.country = data


class CountryEntity(object):
    """Represents a country database entity"""

    def __init__(self, data):
        self.abbreviation,\
            self.name = data


class StreetSuffixEntity(object):
    """Represents a street suffix entity"""

    def __init__(self, data):
        self.suffix_name = data[0]



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from cahoots.parsers.base import BaseParser
from cahoots.util import truncate_text
from cahoots.config import BaseConfig
from cahoots.confidence.normalizer import HierarchicalNormalizerChain
import datetime
import threading
import time
import inspect


class ParserThread(threading.Thread):
    """Represents a thread that will handle one parser parsing request"""
    config = None
    data_string = None
    results = []

    def __init__(self, config, module, data_string):
        self.config = config
        self.thread_id = module
        self.data_string = data_string
        threading.Thread.__init__(self)

    def run(self):
        parser = self.thread_id(self.config)
        self.results = parser.parse(self.data_string) or []


class CahootsParser(object):
    '''Kicks off the parsing process'''

    config = None

    # The 'config' variable, if set, needs to be a class that extends
    # BaseConfig, or an instance of a class that does.
    # In the case that it's a class, we will instantiate the class.
    def __init__(self, config=None, bootstrap=False):

        if config is not None:
            if inspect.isclass(config) and issubclass(config, BaseConfig):
                self.config = config()
            elif isinstance(config, BaseConfig):
                self.config = config

        # Config fallback
        if self.config is None:
            self.config = BaseConfig()

        # This bootstraps our parsing system and
        # gets all modules ready for parsing.
        if bootstrap:
            self.bootstrap(self.config)

    @classmethod
    def bootstrap(cls, config):
        """Bootstraps each parser. Can be used for cache warming, etc."""
        for module in config.enabled_modules:
            # If the module overrides the base bootstrap,
            # we output a message about it
            if module.bootstrap != BaseParser.bootstrap and config.debug:
                print(' * ' + time.strftime('%X %x %Z') +
                      ' * Bootstrapping '+module.__name__)

            module.bootstrap(config)

    def parse(self, data_string):
        """Parses input data and returns a dict of result data"""

        start_time = time.time()
        results = []
        threads = []

        # Creating/starting a thread for each parser module
        for module in self.config.enabled_modules:
            thread = ParserThread(self.config, module, data_string)
            thread.start()
            threads.append(thread)

        # Synchronizing/finishing parser threads
        for thr in threads:
            thr.join()

        # The threads are done, let's get the results out of them
        for thr in threads:
            results.extend(thr.results)

        # Unique list of all major types
        types = list(set([result.type for result in results]))

        if results:
            # Getting a unique list of result types.
            all_types = []
            for res in results:
                all_types.extend([res.type, res.subtype])

            # Hierarchical Confidence Normalization
            normalizer_chain = HierarchicalNormalizerChain(
                self.config,
                types,
                list(set(all_types))
            )
            results = normalizer_chain.normalize(results)

            # Sorting our results by confidence value
            results = sorted(
                results,
                key=lambda result: result.confidence,
                reverse=True
            )

        return {
            'query': truncate_text(data_string),
            'date': datetime.datetime.now(),
            'execution_seconds': time.time() - start_time,
            'top': results[0] if len(results) > 0 else None,
            'results': {
                'count': len(results),
                'types': types,
                'matches': results
            }
        }



"""
The MIT License (MIT)

Copyright (c) Serenity Software, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from SereneRegistry import registry
from cahoots.parsers.base import BaseParser
from cahoots.data import DataHandler
from datetime import timedelta, datetime
import dateutil.parser as dateUtilParser
from pyparsing import\
    Or, \
    CaselessLiteral, \
    ParseException, \
    Word, \
    originalTextFor, \
    ZeroOrMore, \
    nums, \
    alphas, \
    StringEnd


class DateParser(BaseParser):
    '''Determines is given data is a date'''

    @staticmethod
    def bootstrap(config):

        time_scales = [
            'microseconds',
            'milliseconds',
            'seconds',
            'minutes',
            'hours',
            'days',
            'weeks',
            'years',
            'microsecond',
            'millisecond',
            'second',
            'minute',
            'hour',
            'day',
            'week',
            'year',
        ]

        # <number> <timescale> <preposition>
        # 3 seconds until / 50 seconds since
        pre_timedeltas = Or(
            [DateParser.create_pre_timedelta_literal(t) for t in time_scales]
        )
        pre_timedelta_phrases = \
            pre_timedeltas + Word(alphas + nums + " .,;-/'")
        registry.set('DP_pre_timedelta_phrases', pre_timedelta_phrases)

        # <operator> <number> <timescale>
        # plus 5 hours / - 17 days
        post_timedelta_phrases = Or(
            [DateParser.create_post_timedelta_literal(t) for t in time_scales]
        )
        registry.set('DP_post_timedelta_phrases', post_timedelta_phrases)

    @staticmethod
    def create_pre_timedelta_literal(tok):
        """Detects <number> <timescale> <preposition>"""
        delta = originalTextFor(Or([
            Word(nums) +
            ZeroOrMore(',' + Word(nums+',')) +
            ZeroOrMore('.' + Word(nums)),
            CaselessLiteral('an'),
            CaselessLiteral('a')
        ])) + CaselessLiteral(tok) + DateParser.get_preposition_literals()

        delta.setName('pre' + tok).\
            setParseAction(DateParser.generate_pre_timedelta)

        return delta

    @staticmethod
    def generate_pre_timedelta(toks):
        """Generates a timedelta object for a delta-prefix match"""
        minus_prepositions = [
            'until',
            'before',
            'to',
            'from',
        ]

        number, timescale, preposition = toks

        number = DateParser.get_number_value(number)

        if preposition in minus_prepositions:
            number = number * -1

        return DateParser.determine_timescale_delta(timescale, number)

    @staticmethod
    def create_post_timedelta_literal(tok):
        """Detects <plus/minus> <number> <timescale>"""
        delta = Or(
            [CaselessLiteral(t) for t in ['+', '-', 'plus', 'minus']]
        ) + originalTextFor(Or([
            Word(nums) +
            ZeroOrMore(',' + Word(nums+',')) +
            ZeroOrMore('.' + Word(nums)),
            CaselessLiteral('an'),
            CaselessLiteral('a')
        ])) + CaselessLiteral(tok) + StringEnd()

        delta.setName('post' + tok).\
            setParseAction(DateParser.generate_post_timedelta)

        return delta

    @staticmethod
    def generate_post_timedelta(toks):
        """Generates a timedelta object for a delta-suffix match"""
        operator, number, timescale = toks

        number = DateParser.get_number_value(number)

        if operator in ['minus', '-']:
            number = number * -1

        delta = DateParser.determine_timescale_delta(timescale, number)

        return delta

    @staticmethod
    def get_preposition_literals():
        """Generates the prepositions parser and returns it"""
        if registry.test('DP_prepositions'):
            return registry.get('DP_prepositions')

        prepositions = \
            Or([CaselessLiteral(s) for s in DataHandler().get_prepositions()])

        registry.set('DP_prepositions', prepositions)
        return prepositions

    @staticmethod
    def get_number_value(number):
        """Turns a provided number into a proper float"""
        if number in ['a', 'an']:
            number = 1.0
        else:
            number = \
                float("".join([char for char in number if char in nums+'.']))

        return number

    @staticmethod
    def determine_timescale_delta(timescale, number):
        """Gets a timedelta representing the change desired"""
        if timescale[-1:] != 's':
            timescale += 's'

        if timescale == 'microseconds':
            delta = timedelta(microseconds=number)
        elif timescale == 'milliseconds':
            delta = timedelta(milliseconds=number)
        elif timescale == 'seconds':
            delta = timedelta(seconds=number)
        elif timescale == 'minutes':
            delta = timedelta(minutes=number)
        elif timescale == 'hours':
            delta = timedelta(hours=number)
        elif timescale == 'days':
            delta = timedelta(days=number)
        elif timescale == 'weeks':
            delta = timedelta(weeks=number)
        elif timescale == 'years':
            delta = timedelta(days=365*number)
        else:
            delta = timedelta()
        return delta

    def __init__(self, config):
        BaseParser.__init__(self, config, "Date", 0)

    @classmethod
    def natural_parse(cls, data):
        """
        Parse out natural-language strings like "yesterday", "next week", etc
        """
        data = data.lower()
        today = datetime.today()

        if data in ['now', 'current time']:
            value = today.now()
        elif data == 'today':
            value = today
        elif data == "tomorrow":
            value = today + timedelta(1)
        elif data == "yesterday":
            value = today + timedelta(-1)
        elif data == "next week":
            value = today + timedelta(7)
        elif data == "last week":
            value = today + timedelta(-7)
        elif data == "next year":
            value = today + timedelta(365)
        elif data == "last year":
            value = today + timedelta(-365)
        else:
            value = False

        return value

    def date_parse(self, data):
        """Uses the dateUtilParser to determine what our date is"""
        parsed_date = self.natural_parse(data)
        if parsed_date:
            return ('Natural', parsed_date)
        else:
            try:
                return ('Standard', dateUtilParser.parse(data))
            except BaseException:
                pass

        return False

    def parse(self, data_string):
        data_string = data_string.strip()

        if len(data_string) < 3 or len(data_string) > 50:
            return

        # Just date detection
        parsed_date = self.date_parse(data_string)
        if parsed_date:
            yield self.result(parsed_date[0], 100, parsed_date[1])
            return

        # Looking for <number> <timescale> <prepositions> <datetime>
        pre_timedelta_phrases = registry.get('DP_pre_timedelta_phrases')
        try:
            pre_delta = pre_timedelta_phrases.parseString(data_string)
        except ParseException:
            pass
        else:
            parsed_date = self.date_parse(pre_delta[1])
            if parsed_date:
                try:
                    yield self.result(
                        "Number Timescale Preposition Date",
                        100,
                        parsed_date[1] + pre_delta[0]
                    )
                except OverflowError:
                    pass
                return

        # Looking for <datetime> <plus/minus> <number> <timescale>
        post_timedelta_phrases = registry.get('DP_post_timedelta_phrases')
        post_deltas = \
            [t for t in post_timedelta_phrases.scanString(data_string)]
        if len(post_deltas) == 1:
            for token, start, _ in post_deltas:
                parsed_date = self.date_parse(data_string[0:start].strip())
                if parsed_date:
                    try:
                        yield self.result(
                            "Date Operator Number Timescale",
                            100,
                            parsed_date[1] + token.pop()
                        )
                    except OverflowError:
                        pass
                    return
